#include "file_services.h"

int dir_entries_count;
int lcd_cur_line;
int last_visible_entry;
bool entry_selected;

char dir_entries[MAX_ENTRIES][18];
char *curr_dir;

serial *lcd;

int init_fs()
{
    int ret;

    lcd_cur_line = 0;
    last_visible_entry = 3;
    entry_selected = false;    
    ret = sd_mount(SD_DO, SD_CLK, SD_DI, SD_CS);
    lcd = serial_open(LCD_RX, LCD_RX, 0, 19200);
    
    set_direction(LED_YELLOW, 1);
    set_output(LED_YELLOW, 0);
    
    set_direction(LED_RED, 1);
    set_output(LED_RED, 0);

    writeChar(lcd, LCD_ON_CUR_ON_NO_BLK);
    writeChar(lcd, LCD_BACK_LT_ON);
    writeChar(lcd, LCD_ALERT_LENGTH);
    writeChar(lcd, LCD_ALERT_SCALE);
    
    curr_dir = "/";    
    index_dir_entries("/");
    //index_dir_entries("/level_1/level_2/level_3/level_4/level_5");

    update_lcd();

    monitor_ui();
    
    return ret;
}

int create_dir(char *path)
{
    int ret;
    
    ret = mkdir(path, 0);
    
    return ret;
}    

int delete_dir(char *path)
{
    int ret;
    
    ret = rmdir(path);
    
    return ret;
}

FILE *open_file(char *filename, char *mode)
{
    FILE* file_pointer = fopen(filename, mode);
    
    return file_pointer;
}

int close_file(FILE *file_pointer)
{
    int ret;

    ret = fclose(file_pointer);

    return ret;
}

size_t write_file(void *ptr, size_t size, size_t count, FILE *file_pointer)
{
    size_t bytes_written;
    
    bytes_written = fwrite(ptr, size, count, file_pointer);
    
    return bytes_written;
}    

int delete_file(char *filename)
{
    int ret;
    
    ret = remove(filename);
    
    return ret;
}       

char *get_curr_dir(char *buf, int size)
{
    getcwd(buf, size);
    
    return buf;
}    

int set_curr_dir(char *path)
{
    int ret;

    ret = chdir(path);
    
    return ret;
}

char *path_add_slash(char *path)
{
    int path_length;
    char slash = '/';

    if(path)
    {
        path_length = strlen(path);

        if(path[path_length - 1] == slash)
        {
            return &path[path_length];
        }            
        else
        {
            path[path_length] = slash;
            path[path_length + 1] = '\0';

            return &path[path_length + 1];
        }
    }

    return NULL;
}

bool path_append(char *path, char *more)
{
    int path_length;
    char slash = '/';

    if(path && more)
    {
        path_length = strlen(path);

        if(path[path_length - 1] != slash)
        {
            path_add_slash(path);
        }            

        strcat(path, more);
    }

    return false;
}

char *path_prev_level(char *path)
{
    char *pos = strrchr(path, '/');
    
    if (pos != NULL)
    {
        *pos = '\0';
    }
    
    if(strlen(path) == 0)
    {
        return "/";
    }        

    return path;
}

void sort()
{
    bool incomplete = true;
    
    while(incomplete)
    {
        incomplete = false;

        for(int pos = 0; pos < dir_entries_count; pos++)
        {
            incomplete |= compare_swap(pos);
        }
    }
}

bool compare_swap(int pos)
{
    char temp[18];
      
    int pos_first = pos;   
    int pos_next = pos + 1;

    if(strcmp(dir_entries[pos_first], dir_entries[pos_next]) > 0)  
    {
        strcpy(temp, dir_entries[pos_first]);
        memset(dir_entries[pos_first], 0, sizeof(dir_entries[pos_first]));      
        strcpy(dir_entries[pos_first], dir_entries[pos_next]);
        memset(dir_entries[pos_next], 0, sizeof(dir_entries[pos_next]));
        strcpy(dir_entries[pos_next], temp);      

        return true;   
    }   
   
    return false;
}

void empty_dir_entries()
{
    for(int i = 0; i < MAX_ENTRIES; i++)
    {
        clear_entry_str(dir_entries[i]);
    }
}

void clear_entry_str(char *entry)
{
    int i = 0;

    while(entry[i] != '\0')
    {
        entry[i] = '\0';
        i++;
    }
}    

int index_dir_entries(char *dir_path)
{
    empty_dir_entries();
    
    dir_entries_count = -1;

    DIR *sd_dir;
    struct dirent *dir_entry;

    sd_dir = opendir(dir_path);

    if(sd_dir != NULL)
    {
        while((dir_entry = readdir(sd_dir)))
        {
            if((strcmp(dir_entry->d_name, ".") != 0) && 
                (strcmp(dir_entry->d_name, "..") != 0))
            {
                dir_entries_count++;
            
                if(dir_entry->attr == ATTR_DIRECTORY)
                {
                    strcpy(dir_entries[dir_entries_count], " (DIR)");
                    strcat(dir_entries[dir_entries_count], dir_entry->d_name);
                }
                else
                {
                    strcpy(dir_entries[dir_entries_count], " ");
                    strcat(dir_entries[dir_entries_count], dir_entry->d_name);
                }
            }                                
        }        

        closedir(sd_dir);
        
        sort();
    }
    else
    {
        print("opendir could not open %s", dir_path);
    }        

    return 0;
}

void print_dir_entries()
{
    for(int i = 0; i != dir_entries_count + 1; i++)
    {
        print("%s\n", dir_entries[i]);
    }
}

void monitor_ui()
{
    bool item_selected;
    item_selected = false;
    
    while(item_selected == false)
    {
        if(input(LCD_SCR_UP) == 1)
        {
            waitcnt(CNT + CLKFREQ / 4);
            on_scroll_up_lcd();
        }
        
        if(input(LCD_SCR_DWN) == 1)
        {
            waitcnt(CNT + CLKFREQ / 4);
            on_scroll_down_lcd();
        }
        
        if(input(LCD_SELECT) == 1)
        {
            waitcnt(CNT + CLKFREQ / 4);
            on_sel_file_dir();
        }
        
        if(input(LCD_CONFIRM) == 1)
        {
            waitcnt(CNT + CLKFREQ / 4);
            on_confirm_sel();
        }
        
        if(input(LCD_PARENT) == 1)
        {
            waitcnt(CNT + CLKFREQ / 4);
            on_up_level();
        }
        
        if(input(LCD_RESET) == 1)
        {
            waitcnt(CNT + CLKFREQ / 4);
            on_reset_root();
        }
    }        
}

void up_dir_level()
{
    char current_dir[MAX_PATH];
    char *parent_dir;
    
    get_curr_dir(current_dir, MAX_PATH);
    parent_dir = path_prev_level(current_dir);
    set_curr_dir(parent_dir);    
}

void update_lcd()
{
    writeChar(lcd, LCD_CLEAR);

    pause(5);

    dprint(lcd, dir_entries[last_visible_entry - 3]);
    writeChar(lcd, LCD_MV_CUR_LN_1_POS_0);
    dprint(lcd, dir_entries[last_visible_entry - 2]);
    writeChar(lcd, LCD_MV_CUR_LN_2_POS_0);
    dprint(lcd, dir_entries[last_visible_entry - 1]);
    writeChar(lcd, LCD_MV_CUR_LN_3_POS_0);
    dprint(lcd, dir_entries[last_visible_entry]);

    update_cursor();
}

void sound_alert()
{
    writeChar(lcd, LCD_ALERT_NOTE);
}    

void update_cursor()
{
    if(lcd_cur_line == 0)
    {
        writeChar(lcd, LCD_MV_CUR_LN_0_POS_0);
    }
    else if(lcd_cur_line == 1)
    {
        writeChar(lcd, LCD_MV_CUR_LN_1_POS_0);        
    }
    else if(lcd_cur_line == 2)
    {
        writeChar(lcd, LCD_MV_CUR_LN_2_POS_0);
    }
    else if(lcd_cur_line == 3)
    {
        writeChar(lcd, LCD_MV_CUR_LN_3_POS_0);
    }
}

void on_scroll_up_lcd()
{
    clear_selected_entry();
    
    if(lcd_cur_line == 3)
    {
        writeChar(lcd, LCD_MV_CUR_LN_2_POS_0);
        lcd_cur_line = 2;
    }
    else if(lcd_cur_line == 2)
    {
        writeChar(lcd, LCD_MV_CUR_LN_1_POS_0);
        lcd_cur_line = 1;
    }
    else if(lcd_cur_line == 1)
    {
        writeChar(lcd, LCD_MV_CUR_LN_0_POS_0);
        lcd_cur_line = 0;
    }
    else if(lcd_cur_line == 0)
    {
        if(last_visible_entry - 3 != 0)
        {
            last_visible_entry--;
            update_lcd();
        }
        else
        {
            set_output(LED_RED, 1);
            pause(250);
            set_output(LED_RED, 0);
            sound_alert();
        }            
    }
}
    
void on_scroll_down_lcd()
{
    clear_selected_entry();
    
    if(lcd_cur_line == 0)
    {
        writeChar(lcd, LCD_MV_CUR_LN_1_POS_0);
        lcd_cur_line = 1;
    }
    else if(lcd_cur_line == 1)
    {
        writeChar(lcd, LCD_MV_CUR_LN_2_POS_0);        
        lcd_cur_line = 2;
    }
    else if(lcd_cur_line == 2)
    {
        writeChar(lcd, LCD_MV_CUR_LN_3_POS_0);
        lcd_cur_line = 3;
    }
    else if(lcd_cur_line == 3)
    {
        if(last_visible_entry != dir_entries_count)
        {
            last_visible_entry++;
            update_lcd();
        }
        else
        {
            set_output(LED_RED, 1);
            pause(250);
            set_output(LED_RED, 0);
            sound_alert();
        }            
    }
}
    
void on_sel_file_dir()
{
    entry_selected = true;
    
    writeChar(lcd, LCD_ON_CUR_ON_BLK);
}

void clear_selected_entry()
{
    if(entry_selected == true)
    {
        entry_selected = false;
        
        writeChar(lcd, LCD_ON_CUR_ON_NO_BLK);
    }        
}    
    
void on_confirm_sel()
{
    if(entry_selected == true)
    {
        entry_selected = false;

        char *temp = NULL;
    
        if(lcd_cur_line == 0)
        {
            strcpy(temp, dir_entries[last_visible_entry - 3]);
        }
        else if(lcd_cur_line == 1)
        {
            strcpy(temp, dir_entries[last_visible_entry - 2]);
        }
        else if(lcd_cur_line == 2)
        {
            strcpy(temp, dir_entries[last_visible_entry - 1]);
        }
        else if(lcd_cur_line == 3)
        {
            strcpy(temp, dir_entries[last_visible_entry]);
        }
    
        temp++;
    
        if(strncmp(temp, "(DIR)", 5) == 0)
        {
            for(int i = 0; i < 5; i++)
            {
                temp++;
            }
            
            path_append(curr_dir, temp);
            
            // print added for test purposes only
            // remove later            
            print("%s\n", curr_dir);
            
            writeChar(lcd, LCD_ON_CUR_ON_NO_BLK);
    
            index_dir_entries(curr_dir);
    
            lcd_cur_line = 0;    
            last_visible_entry = 3;
    
            update_lcd();
        }
        //add else to indicate a file was chosen
    }    
}
    
void on_reset_root()
{
    clear_selected_entry();
    
    curr_dir = "/";
    index_dir_entries("/");
    
    lcd_cur_line = 0;    
    last_visible_entry = 3;
    
    update_lcd();
}
    
void on_up_level()
{
    clear_selected_entry();
    
    set_output(LED_YELLOW, 1);
    pause(1000);
    set_output(LED_YELLOW, 0);
    
    char *parent_dir;
    
    parent_dir = path_prev_level(curr_dir);
    curr_dir = parent_dir;

    index_dir_entries(curr_dir);
    
    lcd_cur_line = 0;    
    last_visible_entry = 3;    
    
    update_lcd();
}